/*
 * Detect a Library for hardware detection
 *
 * Copyright (C) 1998-2000 MandrakeSoft
 * Copyright (C) 2005 Alexandre Dussart
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef MOUSE_DETECT_SERIAL

# define _GNU_SOURCE

#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <termios.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>


#include "serial.h"
#include "detect.h"

#define NUMBER_OF_VALID_SERIAL_PORTS 5
#define SERIAL_DEVICES "/dev/ttyS"


/* Copyrigt (c) RedHat Sotfware */
/* Modified by bero and alex */

typedef struct INRPAIR{
	int kernel_id;
	int speed;
}IntPair;

int open_serial_port(char *port){
	int fd;
	fd = open(port, O_RDWR | O_NONBLOCK);
	if(fd >= 0)
		if(fcntl(fd, F_SETFL, 0) < 0){
			close(fd);
			return -1;
		}
	return fd;
}

int close_serial_port(int fd){
	return close(fd);
}

int get_serial_lines(int fd){
	int modem_lines;
	ioctl(fd, TIOCMGET, &modem_lines);
	return modem_lines;
}

int set_serial_lines(int fd, int modem_lines){
	return ioctl(fd, TIOCMSET, &modem_lines);
}

int get_serial_attr(int fd, struct termios *attr){
	return tcgetattr(fd, attr);
}

int set_serial_attr(int fd, struct termios *attr){
	return tcsetattr(fd, TCSANOW, attr);
}

int setup_serial_port(int fd, int nbits, struct termios *attr){
	attr->c_iflag = IGNBRK | IGNPAR;
	attr->c_cflag = 0;
	attr->c_cflag &= ~(CSIZE | CSTOPB | PARENB | PARODD | PARENB);
	attr->c_cflag |= CREAD | CLOCAL;
	if(nbits == 7){
		attr->c_cflag |= CS7 | CSTOPB;
	}else{
		attr->c_cflag |= CS8;
	}/*endif*/
	attr->c_oflag = 0;
	attr->c_lflag = 0;

	attr->c_cc[VMIN] = 1;
	attr->c_cc[VTIME] = 5;

	cfsetospeed(attr, B1200);
	cfsetispeed(attr, B1200);
	return set_serial_attr(fd, attr);
}

/**********************************************************************/
/* wait_for_input:  Wait for a reaction on a device attached to a     */
/*                  file descriptor.                                  */
/*                                                                    */
/*     fd: file descriptor.                                           */
/*     timeout: maximum time to wait.                                 */
/**********************************************************************/
int wait_for_input(int fd, struct timeval *timeout){
  fd_set ready;
  int n;
  
  FD_ZERO(&ready);
  FD_SET(fd, &ready);
  
  n = select(fd + 1, &ready, NULL, &ready, timeout);
  
  return n;
}/*endfunc wait_for_input*/




/* Copyright (c) RedHat Sotfware */
/* Dougal: make this external and run from serial_detect*/
int find_legacy_mouse(int fd, int serial_attr, struct termios *portattr){
	int modem_lines, temp;
	unsigned char resp[2];
	struct timeval timo;

	if(serial_attr < 0)
		return 1;  // 1== io problem, nothing to communicate with.
	temp = setup_serial_port(fd, 7, portattr);
	if(temp < 0)
		return 1;  // 1== io problem, nothing to communicate with.

	modem_lines = get_serial_lines(fd);
	modem_lines &= ~(TIOCM_RTS | TIOCM_DTR);
	set_serial_lines(fd, modem_lines);
	usleep(200000);  // sleep 0.2 seconds

	modem_lines |= TIOCM_DTR | TIOCM_RTS;
	set_serial_lines(fd, modem_lines);

	timo.tv_sec = 0;
	timo.tv_usec = 300000;
	if(wait_for_input(fd, &timo) > 0){
		temp = read(fd, resp, 1);
		if(temp < 0){
			if(errno != EAGAIN)
				return 1;   // 1== io problem, nothing to communicate with.
		}
	}
	if(*resp == 'M')
		return 3;   // 3== found mouse
	return 2;   // 2== its a something, but not a mouse as far as we know
}

#ifndef FALSE
#define FALSE 0
#define TRUE (!FALSE)
#endif

int serial_is_unknown(int ttySN){
	int ret=TRUE, found=FALSE;
	FILE *f;
	char *line=NULL;
	size_t len = 0;

	if ((f = fopen("/proc/tty/driver/serial", "r"))){
		while (getline(&line, &len, f) >= 0) {
			if (line[1] != ':')
				continue;
			if (atoi(line) != ttySN)
				continue;
			if (debug)
				printf("\t/dev/ttyS%s", line);
			if (len >= 9 && strncmp(line+8,"unknown",7)!=0)
				ret=FALSE;
			found = TRUE;
			break;
		}
		if (line) free(line);
		fclose(f);
		if (found==0) {
			if (debug) printf("\t/dev/ttyS%d: is UNKNOWN by the kernel.\n", ttySN);
			return TRUE;
		}
	}
	return ret;
}

int ensure_dev_node_present( char *devnode, int maj, int mnr){
	struct stat statfd;
	int mkn, st, create = 0;
	if(debug) printf("\t\tensure_dev_node_present on %s %d %d\n", devnode, maj, mnr);
	if (maj<=0 || mnr<=0)
		return FALSE;
	st = stat(devnode, &statfd); 
	if (st == -1){
		if (errno == ENOENT)
			create = 1;
		else
			if(debug) printf("\t\tstat %s failed %d:%s\n", devnode, errno, strerror(errno));
	}
	if (create){
		if (debug) printf("\t\tCreating node %s major=%d minor=%d\n",devnode, maj, mnr);
		mkn = mknod(devnode, S_IFCHR, (maj*0x100)+(mnr&0xFF));
		if (mkn==-1){
			if (debug) printf("\t\tmknod failed %d:%s\n", errno, strerror(errno));
			return FALSE;
		}
	}
	return TRUE;
}

char *result123(int test){
	if (test==1) return "io error, or nothing to talk to";
	if (test==2) return "unexpected result, its no mouse.";
	if (test==3) return "positive confirmation";
	return "unknown or unexpected representation value (bug)";
}

void serial_detect(void){
	int i;
	int fd;
	int test;
	int serial_attr;
	struct termios origattr;
	char testing_port[50];

	if(debug) fprintf(stdout, "\nProbing serial ports...\n");

	/* Need to try to open /dev/ttyS<i> ports */
	for(i=0; i < NUMBER_OF_VALID_SERIAL_PORTS; i++){
		snprintf(testing_port,49, "%s%d", SERIAL_DEVICES, i);
		if (serial_is_unknown(i)){
			if (force_full_scan==0){
				if(debug) printf("\t\tignore: [%s] (force scan possible with -f)\n", testing_port);
				continue;
			}
			if(debug) printf("\t\tforce a scan anyway: [%s]\n", testing_port);
		}
		if (!ensure_dev_node_present(testing_port, 4, 64+i))
			continue;
		if((fd = open_serial_port(testing_port)) < 0){
			if(debug) printf("\t\tport open fails: %d:%s\n", errno, strerror(errno));
			continue;
		}
		errno=0;
		serial_attr = get_serial_attr(fd, &origattr);
		if(serial_attr < 0){
			if(debug) printf("\t\tNo serial line: %d:%s\n", errno, strerror(errno));
			close_serial_port(fd);
			continue;
		}
		test = find_legacy_mouse(fd, serial_attr, &origattr);

		if (debug) printf("\t\tSerial-Mouse test result is: %s\n",result123(test));
		if(test == 3){  // Dougal: we found a mouse //
			printf("Type:SERIAL-mouse|Port:%s\n", testing_port);
		}else{
			if(debug || show_unpluged)
				printf("Type:SERIAL-Port|Port:%s\n", testing_port);
		}
		close_serial_port(fd);
	}
}

#endif /* MOUSE_DETECT_SERIAL */
